#include "kernel/types.h"
#include "kernel/stat.h"
#include "user/user.h"
#include "kernel/signal.h"


static int failed = 0;
int failedprivate;

void sighuphandler(int signum){
	fprintf(1, "In user defined handler.\n");
}

void sigint_handler(int signum){
	fprintf(1, "Signal delivered, currently running user-defined handler\n");
}


void sigcont_handler(int signum){
	fprintf(1, "Sigcont delivered, in user-defined sigcont handler\n");
	fprintf(1, "Passed\n");
}

void sigusr1_handler(int signum){
	fprintf(1, "Currently in user defined handler. About to call systemcall sleep(2)\n");
	sleep(2);
	fprintf(1, "Back in user defined handler. Context restored properly\n");
}


int blocktest(){
	fprintf(1, "\nTesting sigblock() and sigunblock()\n\n");
	uint mask = siggetmask(), ret = 0;
	int pid = fork();
	
	if(pid == 0){
	
		fprintf(1, "Test 1 : testing sigblock()\n");
		if(sigblock(SIGTERM) == (mask | (1 << SIGTERM)))
			raise(SIGTERM);
		else{
			fprintf(2, "Failed\n");
			sigsetmask(mask);
			exit(0);
		}
		fprintf(1, "Passed\n");
		
		fprintf(1, "Test 2 : testing sigunblock()\n");
		if(sigunblock(SIGTERM) == mask)
			fprintf(1,"Passed\n");
		else{
			fprintf(2, "Failed\n");
			sigsetmask(mask);
			exit(0);
		}
	}
	else{
		wait(0);
		fprintf(1, "Passed\n");
		
		fprintf(1, "Test 3 : block sigkill\n");
		
		if(sigblock(SIGKILL)){
			fprintf(2, "Failed");
			ret++;
		}
		else
			fprintf(1, "Passed\n");
			
		fprintf(1, "Test 4 : block sigstop\n");
		
		if(sigblock(SIGSTOP)){
			fprintf(2, "Failed");
			ret++;
		}
		else
			fprintf(1, "Passed\n");
			
		fprintf(1, "Test 5 : testing validation of input\n");
		
		if(sigblock(0) == -1)
			fprintf(1, "5.1 Passed; ");
		else
			fprintf(2, "5.1 Failed; ");
			
		if(sigblock(-24) == -1)
			fprintf(1, "5.2 Passed; ");
		else
			fprintf(2, "5.2 Failed; ");
			
		if(sigblock(300) == -1)
			fprintf(1, "5.3 Passed; ");
		else
			fprintf(2, "5.3 Failed; ");
			
		if(sigblock(25) == -1)
			fprintf(1, "5.4 Passed; ");
		else
			fprintf(2, "5.4 Failed; ");	
				
		return ret;
		
	}

	return -1;
}




int Killtest(){
	int pid = getpid();
	
	fprintf(1, "\nTesting systemcall Kill()\n");
	fprintf(1, "\nTest 1 : Testing a signal that is ignored by default - SIGCHLD\n");
	
	Kill(pid, SIGCHLD);
	fprintf(1, "Passed\n");
	fprintf(1, "Test 2 : Testing a signal that causes a process to terminate by default\n");
	
	pid = fork();
	if(pid == 0){
		raise(SIGINT);
		return 1;
	}
	else{
		wait(0);
		fprintf(1, "Passed\n");
	}
	
	fprintf(1, "Test 3 : Testing a signal that causes a process to dump core by default\n");
	pid = fork();
	if(pid == 0){
		raise(SIGQUIT);
		return 1;
	}
	else{
		wait(0);
		fprintf(1, "Passed\n");
	}
	
	fprintf(1, "Test 4 : Testing SIGCONT and a signal which causes a process to stop\n");
	pid = fork();
	if(pid == 0){
		int i = 0;
		while(i < 50)
			i++;
		exit(0);
		
	}
	else{
		
		Kill(pid, SIGTSTP);
		fprintf(1, "Stopped process with pid %d\n", pid);
		fprintf(1, "Resuming process with pid %d\n", pid);
		Kill(pid, SIGCONT);
		Kill(pid, SIGKILL);
		wait(0);
		fprintf(1, "Passed\n");
	}
	
	fprintf(1, "Test 5 : Trying to kill init\n");
	Kill(1, SIGKILL);
	fprintf(1, "Passed\n");
	
	fprintf(1, "Test 6 : Mask a signal, and see if it is delivered\n");
	pid = fork();
	if(pid == 0){
	
		int mask = siggetmask();
		sigsetmask(mask | (1 << SIGINT));
		signal(SIGINT, sigint_handler);
		Kill(getpid(), SIGINT);
		fprintf(1, "Passed\n");
		exit(0);
	}
	else{
		wait(0);
		
	}
	
	fprintf(1, "Test 7 : Giving Kill invalid signal values\n");
	if(Kill(getpid(), 0) == -1)
		fprintf(1, "7.1 Passed; ");
	else
		fprintf(2, "7.1 Failed; ");
		
	if(Kill(getpid(), -24) == -1)
		fprintf(1, "7.2 Passed; ");
	else
		fprintf(2, "7.2 Failed; ");
		
	if(Kill(getpid(), 300) == -1)
		fprintf(1, "7.3 Passed; ");
	else
		fprintf(2, "7.3 Failed; ");
		
	if(Kill(getpid(), 25) == -1)
		fprintf(1, "7.4 Passed; ");
	else
		fprintf(2, "7.4 Failed; ");	
		
	return 0;
}


int testsignal_syscall(int pid){
	fprintf(1, "\nTesting syscall signal\n\n");
	int pidtwo;
	failedprivate = 0;
	pidtwo = fork();
	if(pidtwo == 0){
		fprintf(1, "Test 1 : Checking SIG_IGN\n");
		signal(SIGHUP, SIG_IGN);
		Kill(getpid(), SIGHUP);
		fprintf(1, "Passed\n");
		fprintf(1, "Test 2 : Checking user defined handler\n");
		
		
		signal(SIGHUP, sighuphandler);
		Kill(getpid(), SIGHUP);
		fprintf(1, "Passed\n");
		fprintf(1, "Test 3 : Checking SIG_DFL\n");
		signal(SIGHUP, SIG_DFL);
		Kill(getpid(), SIGHUP);
		return 0;
		
	}
	else{
		wait(0);
		fprintf(1, "Passed\n");
		
	}
		
	fprintf(1, "Test 4 : Trying to handle SIGKILL\n");
	if(signal(SIGKILL, sighuphandler) == SIG_ERR)
		fprintf(1, "Passed\n");
	
		
	fprintf(1, "Test 5 : Trying to handle SIGSTOP\n");
	if(signal(SIGSTOP, sighuphandler) == SIG_ERR)
		fprintf(1, "Passed\n");
	
	// fprintf(1, "Test 8 : User defined handler calling system calls\n");
	// signal(SIGUSR1, sigusr1_handler);
	// Kill(getpid(), SIGUSR1);
	// fprintf(1, "Passed\n");
	// signal(SIGUSR1, SIG_DFL);
	
	fprintf(1, "Test 6 : Giving invalid values as arguments to signal\n");
	
	if(signal(0, SIG_DFL) == SIG_ERR)
		fprintf(1, "6.1 Passed; ");
	else
		fprintf(2, "6.1 Failed; ");
		
	if(signal(-24, SIG_DFL) == SIG_ERR)
		fprintf(1, "6.2 Passed; ");
	else
		fprintf(2, "6.2 Failed; ");
		
	if(signal(300, SIG_DFL) == SIG_ERR)
		fprintf(1, "6.3 Passed; ");
	else
		fprintf(2, "6.3 Failed; ");
		
	if(signal(25, SIG_DFL) == SIG_ERR)
		fprintf(1, "6.4 Passed; ");
	else
		fprintf(2, "6.4 Failed; ");
	
	
	fprintf(1, "\nTest 7 : Testing SIGCONT handlers\n");	
	signal(SIGCONT, sigcont_handler);
	raise(SIGCONT);
	signal(SIGCONT, SIG_DFL);
	fprintf(1, "Passed\n");
	return 1;
	
}


int masktest(){
	uint mask;
	int fail = 0;
	mask = ((1 << SIGHUP) | (1 << SIGINT));		//block sighup and sigint		
	fprintf(1, "\ntesting sigsetmask and siggetmask\n\n");
	fprintf(1, "Test 1 : checking normal functionality\n");
	sigsetmask(mask);
	if(siggetmask() == mask)
		fprintf(1, "Passed\n");
	else{
		fprintf(2, "Failed\n");
		fail++;
	}
	
	fprintf(1, "Test 2 : SIGKILL\n");
	mask = 0;
	mask = (1 << SIGKILL);
	sigsetmask(mask);
	if(siggetmask() == mask){
		fprintf(2, "Failed\n");
		fail++;
	}
	else{
		fprintf(1, "Passed\n");
	
	}
	
	fprintf(1, "Test 3 : SIGSTOP\n");
	mask = 0;
	mask = (1 << SIGSTOP);
	sigsetmask(mask);
	if(siggetmask() == mask){
		fprintf(2, "Failed\n");
		fail++;
	}
	else{
		fprintf(1, "Passed\n");
	
	}
	
	fprintf(1, "Test 4 : SIGCONT\n");
	mask = 0;
	mask = (1 << SIGCONT);
	sigsetmask(mask);
	if(siggetmask() == mask){
		fprintf(2, "Failed\n");
		fail++;
	}
	else{
		fprintf(1, "Passed\n");
	
	}
	
	sigsetmask(0);
	return fail;
}



int testsigkill(){
	int pid;
	fprintf(1, "\ntesting SIGKILL\n\n");
	pid = fork();
	if(pid == 0){
		while(1)
			;
		exit(0);
		
	}
	else{
		sleep(1);
		Kill(pid, SIGKILL);
		wait(0);
		fprintf(1, "Sigkill delivered, passed.\n");
		return 0;
	}
	return -1;
}


int testsigterm(){
	int pid;
	fprintf(1, "\ntesting SIGTERM\n\n");
	pid = fork();
	if(pid == 0){
		while(2)
			;
		exit(0);
	}
	else{
		sleep(1);
		Kill(pid, SIGTERM);
		wait(0);
		fprintf(1, "Sigterm delivered, passed.\n");
		return 0;
	}
	return -1;
}


int main(){
	
	int ret, pid = getpid();
	fprintf(1, "\nRUNNING TESTSUITE FOR SIGNALS\n");
	fprintf(1, "\n=================================\n");
	if(testsigkill() == -1){
		fprintf(2, "testsigkill failed\n");
		failed++;
	}
	
	if(testsigterm() == -1){
		fprintf(2, "testsigterm failed\n");
		failed++;
	}
	
	fprintf(1, "\n=================================\n");
	ret = masktest();
	if(ret != 0){
		fprintf(2, "masktest failed\n");
		failed = failed + ret;
	}
	fprintf(1, "\n=================================\n");
	ret = blocktest();
	if(ret != 0){
		fprintf(2, "blocktest failed\n");
		failed = failed + ret;
	}
	fprintf(1, "\n\n=================================\n");
	ret = Killtest();
	if(ret != 0){
		fprintf(2, "Killtest failed\n");
		failed = failed + ret;
	}
	fprintf(1, "\n\n=================================\n");
	if(!testsignal_syscall(pid)){
		fprintf(2, "syscall signal() test failed\n");
		failed++;
		exit(0);
	}
	fprintf(1, "\n=================================\n");
	if(failed == 0)
		fprintf(1, "\nAll tests passed\n");
	else
		fprintf(2, "%d tests failed partially \n", failed);
	
	exit(0);
}
